/*
 *	functions for the graphic output of the pixmap
 */


#include "chemtool.h"


void CreatePix()
{
    XGCValues values;
    XtGCMask mask = GCForeground | GCBackground | GCLineStyle | GCFunction;

    float ztab[]={ 0.6, 0.8, 1};
    int ltab[]={ 0, 2, 2};
    int ftab[]={ 3, 5, 6};

    size_factor=ztab[zoom_factor];
    head.pix_width=head.width*size_factor;
    head.pix_height=head.height*size_factor;

    picture = XCreatePixmap(XtDisplay(pix),
	RootWindowOfScreen(XtScreen(pix)),
	head.pix_width, head.pix_height, 1); 

    values.foreground=0;
    values.background=1;
    values.line_style=LineSolid;
    values.function=GXcopy;
    cleargc = XCreateGC(XtDisplay(pix), picture, mask, &values);

    values.foreground=1;
    values.background=0;
    values.line_style=LineSolid;
    values.function=GXcopy;
    drawgc =  XCreateGC(XtDisplay(pix), picture, mask, &values);

    values.foreground=1;
    values.background=0;
    values.line_style=LineSolid;
    values.function=GXinvert;
    invgc = XCreateGC(XtDisplay(pix), picture, mask, &values);

    if(!Load_Font(ftab[zoom_factor])){
    if( (font = XLoadQueryFont(XtDisplay(pix), "fixed")) != NULL)
	XSetFont(XtDisplay(pix), drawgc, font->fid);
    else 
	fprintf(stderr, "chemtool: can't load any font\n"); exit(1);
	}
    Set_Line(ltab[zoom_factor]);
    XtVaSetValues(pix, XtNwidth, head.pix_width, 
    XtNheight, head.pix_height, NULL);
}

void FreePix(void)
{
    XFreeGC(XtDisplay(pix), invgc);
    XFreeGC(XtDisplay(pix), drawgc);
    XFreeGC(XtDisplay(pix), cleargc);
    XFreePixmap(XtDisplay(pix), picture);
}

void CopyPlane(void)
{
    XCopyPlane(XtDisplay(pix), picture, XtWindow(pix), 
    DefaultGCOfScreen(XtScreen(pix)), 
    0, 0, head.pix_width, head.pix_height, 0, 0, 1);
}

void Display_Mol(void)
{
    struct data *hpc;
    struct dc *hpc_c;
    int d;
/*    int xb,yb,d;*/
    char *c;
    struct xy_co *coord;

    XFillRectangle(XtDisplay(pix), picture, cleargc, 0, 0, 
		      head.pix_width, head.pix_height);

     hpc=da_root.next;
     for(d=0; d<hp.n; d++){
	  if(!hpc->bond) Drawline(hpc->x, hpc->y, hpc->tx, hpc->ty); /*single*/
          if(hpc->bond==5) DrawWedge(hpc->x, hpc->y, hpc->tx, hpc->ty);
          if(hpc->bond==6) DrawDashedWedge(hpc->x, hpc->y, hpc->tx, hpc->ty);
          if(hpc->bond==7) DrawWiggly(hpc->x, hpc->y, hpc->tx, hpc->ty);
          if(hpc->bond==8) DrawArrow(hpc->x, hpc->y, hpc->tx, hpc->ty,1);
          if(hpc->bond==9) DrawArrow(hpc->x, hpc->y, hpc->tx, hpc->ty,2);
	  if(hpc->bond==4){ /*middouble*/
	      coord=center_double_bond(hpc->x, hpc->y, hpc->tx, hpc->ty, 4);
	      Drawline(coord->x, coord->y, coord->tx, coord->ty);
	      coord++;
	      Drawline(coord->x, coord->y, coord->tx, coord->ty);
	      }
	  if(hpc->bond==1 || hpc->bond==3){ /*leftdouble, triple*/
	      coord=multi_bonds(hpc->x, hpc->y, hpc->tx, hpc->ty, 10);
	      Drawline(coord->x, coord->y, coord->tx, coord->ty);
	      Drawline(hpc->x, hpc->y, hpc->tx, hpc->ty);
	      }
	  if(hpc->bond==2 || hpc->bond==3){ /*rightdouble, triple*/
	      coord=multi_bonds(hpc->tx, hpc->ty, hpc->x, hpc->y, 10);
	      Drawline(coord->x, coord->y, coord->tx, coord->ty);
	      Drawline(hpc->x, hpc->y, hpc->tx, hpc->ty);
	      }
	   hpc=hpc->next; 
    }

    hpc_c=dac_root.next;
    for(d=0; d<hp.nc; d++){
	 c=hpc_c->c;
	 Drawstring(hpc_c->x, hpc_c->y, c, hpc_c->direct);
	 hpc_c=hpc_c->next; 
	 }

    if(mark.flag){
	 XDrawRectangle(XtDisplay(pix), picture, drawgc, 
	 mark.x*size_factor, mark.y*size_factor, 
	 mark.w*size_factor, mark.h*size_factor);
    }
    CopyPlane();
}


void Drawline(int x, int y, int tx, int ty)
{
    XDrawLine(XtDisplay(pix), picture, drawgc,
    x*size_factor, y*size_factor, tx*size_factor, ty*size_factor);
}

void DrawWiggly(int x, int y, int tx, int ty)
{
struct {
 short x, y;
 unsigned short width, height;
 short angle1, angle2;             /* Degrees * 64 */
}thearcs[5];
int narcs,veclen,boxlen;
int i,xlen,ylen,ang;
int boxcorx,boxcory;
narcs=5;

xlen=tx-x;
ylen=ty-y;

veclen=calc_vector(abs(tx-x),abs(ty-y));
ang=180.+asin(copysign((float)ylen/(float)veclen,xlen*-ylen))*180./3.14159;

boxlen=0.2*veclen;
boxcorx=(1.-(float)xlen/(float)veclen)*boxlen/2;
boxcory=(1.-(float)ylen/(float)veclen)*boxlen/2;

for(i=0;i<narcs;++i){
thearcs[i].x=(x+0.2*i*xlen-boxcorx)*size_factor;
thearcs[i].y=(y+0.2*i*ylen-boxcory)*size_factor;
thearcs[i].width=(unsigned short) boxlen*size_factor;       
thearcs[i].height=(unsigned short) boxlen*size_factor;
thearcs[i].angle1= (short)  ang*64;
thearcs[i].angle2= (short)-180*64*pow(-1,i);

}
/*    XDrawRectangles(XtDisplay(pix), picture, drawgc,(XRectangle*)thearcs,narcs);*/
    XDrawArcs(XtDisplay(pix), picture, drawgc,(XArc*)thearcs,narcs);
/*    XDrawLine(XtDisplay(pix), picture, drawgc,
    x*size_factor, y*size_factor, tx*size_factor, ty*size_factor);*/
}

void DrawArrow(int x, int y, int tx, int ty, int head)
{
int xlen,ylen,xbase,ybase,npoints;
float veclen,headfact;
struct  {
	short x,y;
	} mypoints[3];

xlen=tx-x;
ylen=ty-y;
veclen=sqrt(xlen*xlen+ylen*ylen);
headfact=1.-20./veclen;
xbase=x+headfact*xlen;
ybase=y+headfact*ylen;


mypoints[0].x=tx*size_factor;
mypoints[0].y=ty*size_factor;

mypoints[1].x=(xbase+0.1*ylen)*size_factor;
mypoints[1].y=(ybase-0.1*xlen)*size_factor;
if (head == 1) {
mypoints[2].x=xbase*size_factor; /*on baseline*/
mypoints[2].y=ybase*size_factor;
} else {
mypoints[2].x=(xbase-0.1*ylen)*size_factor;
mypoints[2].y=(ybase+0.1*xlen)*size_factor;
}

/*
fprintf(stderr,"Pfeil  (%d,%d)-(%d,%d)\n",x,y,tx,ty); 
fprintf(stderr,"Spitze  (%d,%d)-(%d,%d)-(%d,%d)\n",
mypoints[0].x,mypoints[0].y,mypoints[1].x,mypoints[1].y,
mypoints[2].x,mypoints[2].y); 
*/

npoints=3;
    XDrawLine(XtDisplay(pix), picture, drawgc,
    x*size_factor, y*size_factor, tx*size_factor, ty*size_factor);
    XFillPolygon(XtDisplay(pix), picture, drawgc,(XPoint*)mypoints,npoints,Convex,CoordModeOrigin);
}


void DrawDashedWedge(int x, int y, int tx, int ty)
{
int i, xlen,ylen;
xlen=tx-x;
ylen=ty-y;

    for  (i= 0; i <11;i++) 
    XDrawLine(XtDisplay(pix), picture, drawgc,
(x+0.1*i*xlen-0.01*ylen*i)*size_factor,(y+0.1*i*ylen+0.01*xlen*i)*size_factor,
(x+0.1*i*xlen+0.01*ylen*i)*size_factor,(y+0.1*i*ylen-0.01*xlen*i)*size_factor);

}

void DrawWedge(int x, int y, int tx, int ty)
{
int xlen,ylen,npoints;
struct  {
	short x,y;
	} mypoints[3];

xlen=tx-x;
ylen=ty-y;

mypoints[0].x=x*size_factor;
mypoints[0].y=y*size_factor;
mypoints[1].x=(tx-0.1*ylen)*size_factor;
mypoints[1].y=(ty+0.1*xlen)*size_factor;
mypoints[2].x=(tx+0.1*ylen)*size_factor;
mypoints[2].y=(ty-0.1*xlen)*size_factor;
npoints=3;
    XFillPolygon(XtDisplay(pix), picture, drawgc,(XPoint*)mypoints,npoints,Convex,CoordModeOrigin);
}

void Drawstring(int x, int y, char *c, int direct)
{
    int ha, htw, tw, a;
    int nch, textl, chl;
    char *l, *hc, hl[2];
    int d, hx, n;
    int variance[MAXCL];
    char text[MAXCL];
    char symbol[MAXCL];

    x=x*size_factor;
    hx=x;
    y=y*size_factor;
    ha=font->ascent/2;
    a=font->ascent;

    n = strlen(c);

    hc=c;
    for(d=0;d<n;d++){
      symbol[d]=FALSE;
      if((variance[d]=Extra_char(*hc) ) != 0) {hc++;--n;}
/*      if(variance[d]==1000)  {variance[d]=0;symbol[d]=TRUE;}*/
      if(variance[d]==1000)  {variance[d]=0;symbol[d]=TRUE;}
      text[d]=*hc++;
      if(text[d]=='@')  {symbol[d]=TRUE;text[d]=*hc++;}
      if (text[d]=='\\') text[d]=' ';
    }
    
    nch=n;
    if(nch>1){
	 if(variance[nch-1]) --nch;
	 textl = XTextWidth(font, c, nch);
	 chl = (textl/nch)/2;
	 hx = x + (textl/2-chl )*direct;
    }
    nch=n;

    x=hx;
    for(d=0;d<n;d++){
	  hl[0]=text[d]; hl[1]=0; l=&hl[0];
	  tw = XTextWidth(font, l, strlen(l)); htw = tw/2;
	  XFillArc(XtDisplay(pix), picture, cleargc, 
			  x-htw - 7*size_factor, 
			  y-a+ha - 5*size_factor + variance[d],
			  htw*2 + 12*size_factor,
			  a + 12*size_factor,
			  0, 360*64);
	  x=x+tw;
    }

    n=nch;
    x=hx;
    for(d=0;d<n;d++){
	  hl[0]=text[d]; hl[1]=0; l=&hl[0];
if (symbol[d]){
	XSetFont(XtDisplay(pix), drawgc, symbfont->fid);
	XSetFont(XtDisplay(pix), cleargc, symbfont->fid);
	}else{
	XSetFont(XtDisplay(pix), drawgc, font->fid);
	XSetFont(XtDisplay(pix), cleargc, font->fid);
	}
	  ha=smallfont->ascent/2;
if (variance[d]) {
if (symbol[d]){
XSetFont(XtDisplay(pix), drawgc, ssymbfont->fid);
XSetFont(XtDisplay(pix), cleargc, ssymbfont->fid);
}else{
XSetFont(XtDisplay(pix), drawgc, smallfont->fid);
XSetFont(XtDisplay(pix), cleargc, smallfont->fid);
}
	  tw = XTextWidth(smallfont, l, strlen(l)); htw = tw/2;
	  ha=smallfont->ascent/2;
}
	XSetFont(XtDisplay(pix), cleargc, font->fid);
	  tw = XTextWidth(font, l, strlen(l)); htw = tw/2;
	  ha=smallfont->ascent/2;
if (variance[d]) {
if (symbol[d]){
XSetFont(XtDisplay(pix), drawgc, ssymbfont->fid);
XSetFont(XtDisplay(pix), cleargc, ssymbfont->fid);
}else{
XSetFont(XtDisplay(pix), drawgc, smallfont->fid);
XSetFont(XtDisplay(pix), cleargc, smallfont->fid);
}
	  tw = XTextWidth(smallfont, l, strlen(l)); htw = tw/2;
	  ha=smallfont->ascent/2;
}
	  XDrawString(XtDisplay(pix), picture, drawgc, 
	  x-htw, y+ha+variance[d], l, strlen(l));
	  x=x+tw;
    }
}

int Extra_char(char c)
{
    int ha=font->ascent/2;
    if(c == '_') return(ha);
    if(c == '^') return(-ha);
/*    if(c == '@') return(1000);*/ /* hack to encode other character set */
    return(0);
}

int Load_Font(int fnt)
{
	
    String symtype[] = {"*-symbol-*-*-8-*", "*-symbol-*-*-10-*", 
    "*-symbol-*-*-12-*", "*-symbol-*-*-14-*", "*-symbol-*-*-18-*", 
    "*-symbol-*-*-20-*", "*-symbol-*-*-24-*"};
    String ssymtype[] = {"*-symbol-*-*-8-*", "*-symbol-*-*-8-*", 
    "*-symbol-*-*-10-*", "*-symbol-*-*-12-*", "*-symbol-*-*-14-*", 
    "*-symbol-*-*-18-*", "*-symbol-*-*-20-*"};

    String fntype[] = {"5x7", "6x10", "7x13", "9x15", "8x16", 
    "10x20", "12x24"};
    String sfntype[] = {"5x7", "5x7", "6x10", "8x13", "7x14", 
    "8x13bold", "10x20"};

    if(fnt>6) return(0);
	
    if((font = XLoadQueryFont(XtDisplay(pix), fntype[fnt])) !=NULL){
	      XSetFont(XtDisplay(pix), drawgc, font->fid);
	      XSetFont(XtDisplay(pix), cleargc,font->fid);
	if ( (symbfont = XLoadQueryFont(XtDisplay(pix),symtype[fnt]))==NULL) symbfont=font;
	if ( (ssymbfont = XLoadQueryFont(XtDisplay(pix),ssymtype[fnt]))==NULL) ssymbfont=font;
	if ( (smallfont = XLoadQueryFont(XtDisplay(pix),sfntype[fnt]))==NULL) smallfont=font;
	      return(1);}
    
    else return(0);
}

void Set_Line(int lsty)
{
    XSetLineAttributes(XtDisplay(pix), gc, lsty, 0, CapButt, JoinMiter);
    XSetLineAttributes(XtDisplay(pix), drawgc, lsty, 0, CapButt, JoinMiter);
    XSetLineAttributes(XtDisplay(pix), cleargc, lsty, 0, CapButt, JoinMiter);
    XSetLineAttributes(XtDisplay(pix), invgc, lsty, 0, CapButt, JoinMiter);
}
