/*
 *	functions for in and output
 */


#include "chemtool.h"
#include "inout.h"

int save_mol(char *filename)
{
    int d;
    struct data *hp_b;
    struct dc *hp_a;
    FILE *fp;

    if((fp = fopen(filename, "w")) == NULL) return(1); 

    fprintf(fp, "Chemtool Version "VERSION"\n");

    fprintf(fp, "geometry %i %i\n", head.width, head.height);

    hp_b=da_root.next;
    fprintf(fp, "bonds %i\n",hp.n);
    for(d=0; d<hp.n; d++){
	 fprintf(fp, "%i	%i	%i	%i	%i\n",
		    hp_b->x, hp_b->y, hp_b->tx, hp_b->ty, hp_b->bond); 
	 hp_b=hp_b->next; 
	}

    hp_a=dac_root.next;
    fprintf(fp, "atoms %i\n",hp.nc);
    for(d=0; d<hp.nc; d++){
	 fprintf(fp, "%i	%i	%s	%i\n",
		    hp_a->x, hp_a->y, hp_a->c, hp_a->direct );
	 hp_a=hp_a->next; 
	}

    fclose(fp);
    return(0);
}

int load_mol(char *filename)
{
    int d,n;
    int x,y,tx,ty,b;
    char str[255];
    FILE *fp;
    char version[10];

    if((fp = fopen(filename, "r")) == NULL) return(1); 

    fscanf(fp, "%s %s %s", str, str, version);
    if(atof(version) > atof(VERSION)) return(2);

    clear_data();

    fscanf(fp, "%s %i %i", str, &x, &y);
    if(!strcmp(str, "geometry")) {
	   head.width=(Dimension) x;
	   head.height=(Dimension) y;
    }

    fscanf(fp, "%s %i", str, &n);
    if(!strcmp(str,"bonds") || !strcmp(str,"bounds")){
	   for(d=0;d<n;d++){
		fscanf(fp, "%i %i %i %i %i",&x,&y,&tx,&ty,&b);
		add_struct(x, y, tx, ty, b);
		}
    }

    fscanf(fp, "%s %i", str, &n);
    if(!strcmp(str,"atoms")){
	   for(d=0;d<n;d++){
		fscanf(fp, "%i %i %s %i",&x,&y,str, &b);
		add_char(x, y, str, b);
		}
    }

    fclose(fp);
    return(0);
}

int add_mol(char *filename)
{
    int d,n;
    int x,y,tx,ty,b;
    int xdiff=0,ydiff=0;
    char str[255];
    FILE *fp;
    char version[10];

    if((fp = fopen(filename, "r")) == NULL) return(1); 

    fscanf(fp, "%s %s %s", str, str, version);
    if(atof(version) > atof(VERSION)) return(2);


    fscanf(fp, "%s %i %i", str, &x, &y);

    fscanf(fp, "%s %i", str, &n);
    if(!strcmp(str,"bonds") || !strcmp(str,"bounds")){
	   for(d=0;d<n;d++){
		fscanf(fp, "%i %i %i %i %i",&x,&y,&tx,&ty,&b);
		if (d==0) {xdiff=mark.x-x;
			   ydiff=mark.y-y; 
			   xref=x+xdiff;
			   yref=y+ydiff;}
		x = x+xdiff;
		y = y+ydiff;
		tx = tx+xdiff;
		ty = ty+ydiff;
		add_struct(x, y, tx, ty, b);
		}
    }

    fscanf(fp, "%s %i", str, &n);
    if(!strcmp(str,"atoms")){
	   for(d=0;d<n;d++){
		fscanf(fp, "%i %i %s %i",&x,&y,str, &b);
		x = x+xdiff;
		y = y+ydiff;
		add_char(x, y, str, b);
		}
    }

    fclose(fp);
    return(0);
}

int export_bitmap(char *filename)
{
    int x, y, w, h;
    Pixmap buffer;
    char xfile[512];

    if(mark.flag){
	x=mark.x*size_factor+3;
	y=mark.y*size_factor+3;
	w=mark.w*size_factor-6;
	h=mark.h*size_factor-6;
    }else{
	x=0; y=0;
	w=head.pix_width; h=head.pix_height;
    }

    buffer = XCreatePixmap(XtDisplay(pix),
		 RootWindowOfScreen(XtScreen(pix)),
		 w, h, 1);

    XCopyPlane(XtDisplay(pix), picture, buffer, drawgc,
	x, y, w, h, 0, 0, 1);


    if(strlen(filename)) {
	    sprintf(xfile, "%s.xbm", filename);
	    XWriteBitmapFile(XtDisplay(pix), xfile, buffer, w, h, -1, -1);
    }
    XFreePixmap(XtDisplay(pix), buffer);
return(0);
}

void xfig_line(FILE *fp, int x, int y, int tx, int ty)
{
    fprintf(fp, "%i %i %i %i %i %i %i %i %i %f %i %i %i %i %i %i\n",
	figline.object, figline.sub_type, figline.line_style,
	figline.thickness, figline.pen_color, figline.fill_color,
	figline.depth, figline.pen_style, figline.area_fill,
	figline.style_val, figline.join_style, figline.cap_style,
	figline.radius, figline.forward_arrow, figline.backward_arrow, 
	figline.npoints);
    fprintf(fp, "%i %i %i %i\n", x, y, tx, ty);
}

void xfig_wedge(FILE *fp, int x, int y, int tx, int ty)
{
    fprintf(fp, "%i %i %i %i %i %i %i %i %i %f %i %i %i %i %i %i\n",
	figline.object, figline.sub_type, figline.line_style,
	figline.thickness, figline.pen_color, figline.fill_color,
	figline.depth, figline.pen_style, figline.area_fill+1,
	figline.style_val, figline.join_style, figline.cap_style,
	figline.radius, figline.forward_arrow, figline.backward_arrow, 
	figline.npoints+1);
    fprintf(fp, "%i %i %i %i %i %i\n", x, y, tx-(int)(0.1*(float)(ty-y)), 
    ty+(int)(0.1*(float)(tx-x)),
    tx+(int)(0.1*(float)(ty-y)), ty-(int)(0.1*(float)(tx-x)));
}

void xfig_wiggly(FILE *fp, int x, int y, int tx, int ty)
{
int arccode=5;
int arctype=1;
int direction=1;
float xlen,ylen,veclen,boxlen;
int boxcorx,boxcory;
int i;
float center_x,center_y;
int xstart,ystart, xmid,ymid,xend,yend;
xlen=tx-x;
ylen=ty-y;

veclen=calc_vector(abs(tx-x),abs(ty-y));
boxlen=0.2*veclen;
boxcorx=(1.-(float)xlen/(float)veclen)*boxlen/2;
boxcory=(1.-(float)ylen/(float)veclen)*boxlen/2;

for(i=0;i<5;i++) {
direction=abs(direction-1);
center_x=x+(0.2*i+.1)*xlen;
center_y=y+(0.2*i+.1)*ylen;
xstart=x+0.2*i*xlen;
ystart=y+0.2*i*ylen;
xmid=center_x-boxcorx*pow(-1,i);
ymid=center_y-boxcory*pow(-1,i);
xend=x+0.2*(i+1)*xlen;
yend=y+0.2*(i+1)*ylen;

    fprintf(fp, "%i %i %i %i %i %i %i %i %i %f %i %i %i %i %f %f %i %i %i %i %i %i\n",
	arccode,arctype,
	figline.line_style,
	figline.thickness, figline.pen_color, figline.fill_color,
	figline.depth, figline.pen_style, figline.area_fill,
	figline.style_val, figline.cap_style,
        direction, 
	figline.forward_arrow, figline.backward_arrow, 
	center_x,center_y,xstart,ystart,xmid,ymid,xend,yend);
  }
}

void xfig_arrow(FILE *fp, int x, int y, int tx, int ty, int type)
{
float xlen,ylen;
float headfact,veclen;
int xbase,ybase,xside,yside,xend,yend;

xlen=tx-x;
ylen=ty-y;
veclen=sqrt(xlen*xlen+ylen*ylen);
headfact=0.95;
xbase=x+headfact*xlen;
ybase=y+headfact*ylen;
xside=xbase+0.1*ylen;
yside=ybase-0.1*xlen;
xend=xbase-type*(0.1*ylen); /*type=0 (half arrow) ends on baseline, */
yend=ybase+type*(0.1*xlen); /*type=1 is a symmetrical arrowhead */
/* the line */
    fprintf(fp, "%i %i %i %i %i %i %i %i %i %f %i %i %i %i %i %i\n",
	figline.object, figline.sub_type, figline.line_style,
	figline.thickness, figline.pen_color, figline.fill_color,
	figline.depth, figline.pen_style, figline.area_fill,
	figline.style_val, figline.join_style, figline.cap_style,
	figline.radius, figline.forward_arrow, figline.backward_arrow, 
	figline.npoints);
    fprintf(fp, "%i %i %i %i\n", x, y, tx, ty);

/* and the arrowhead - a filled polyline */
    fprintf(fp, "%i %i %i %i %i %i %i %i %i %f %i %i %i %i %i %i\n",
	figline.object, figline.sub_type, figline.line_style,
	figline.thickness, figline.pen_color, figline.fill_color,
	figline.depth, figline.pen_style, figline.area_fill+1,
	figline.style_val, figline.join_style, figline.cap_style,
	figline.radius, figline.forward_arrow, figline.backward_arrow, 
	figline.npoints+1);
    fprintf(fp, "%i %i %i %i %i %i\n", tx, ty, xside, 
    yside,xend,yend);
}



int exfig(char *filename, int latex_flag)
{
    struct data *hp_b;
    struct dc *hp_a;
    struct xy_co *coord;

    FILE *fp;
    int i;
    int d, x, y, tx, ty;
    int ha, chl=0;
    int fontsize=12, fzoom=10;
	int xpos,offset,fontshrink;

    switch(zoom_factor){
	case 0:
	    fontsize=6;
	    fzoom=5;
	    break;
	case 1:
	    fontsize=9;
	    fzoom=7;
	    break;
	case 2:
	    fontsize=12;
	    fzoom=10;
	    break;
    }

    if((fp = fopen(filename, "w")) == NULL) return(1);
    fprintf(fp, "#FIG 3.2\n");
    fprintf(fp, "Landscape\n");
    fprintf(fp, "Center\n");
    fprintf(fp, "Inches\n");
    fprintf(fp, "Letter\n");
    fprintf(fp, "100.00\n");
    fprintf(fp, "Single\n");
    fprintf(fp, "0\n");
    fprintf(fp, "1200 2\n");

    figline.object=2;
    figline.sub_type=1;
    figline.line_style=0;
    figline.thickness=1;
    figline.pen_color=0;
    figline.fill_color=7;
    figline.depth=0;
    figline.pen_style=0;
    figline.area_fill=-1;
    figline.style_val=0.000;
    figline.join_style=0;
    figline.cap_style=0;
    figline.radius=-1;
    figline.forward_arrow=0;
    figline.backward_arrow=0;
    figline.npoints=2;

    hp_b=da_root.next;
    for(d=0; d<hp.n; d++){
	coord=bond_cut(hp_b->x, hp_b->y, hp_b->tx, hp_b->ty, 12);
	x=coord->x*fzoom;
	y=coord->y*fzoom;
	tx=coord->tx*fzoom;
	ty=coord->ty*fzoom;
	if(!hp_b->bond) xfig_line(fp, x, y, tx, ty);
	if (hp_b->bond==5) xfig_wedge(fp, x, y, tx, ty);
	if(hp_b->bond==6){
	for (i=0;i<11;i++)
	xfig_line(fp,x+0.1*i*(tx-x)-0.01*(ty-y)*i,
	     y+0.1*i*(ty-y)+0.01*(tx-x)*i,
	     x+0.1*i*(tx-x)+0.01*(ty-y)*i,
	     y+0.1*i*(ty-y)-0.01*(tx-x)*i);
        }	
        if (hp_b->bond==7) xfig_wiggly(fp,x,y,tx,ty);
        if (hp_b->bond==8) xfig_arrow(fp,x,y,tx,ty,0);
        if (hp_b->bond==9) xfig_arrow(fp,x,y,tx,ty,1);
	if(hp_b->bond==4){
	    coord=center_double_bond(x, y, tx, ty, 4*fzoom);
	    xfig_line(fp, coord->x, coord->y, coord->tx, coord->ty);
	    coord++;
	    xfig_line(fp, coord->x, coord->y, coord->tx, coord->ty);
	    }
	if(hp_b->bond==1 || hp_b->bond==3){
	    coord=multi_bonds(x, y, tx, ty, 10*fzoom);
	    xfig_line(fp, x, y, tx, ty);
	    xfig_line(fp, coord->x, coord->y, coord->tx, coord->ty);
	    }
	if(hp_b->bond==2 || hp_b->bond==3){
	    coord=multi_bonds(tx, ty, x, y, 10*fzoom);
	    xfig_line(fp, x, y, tx, ty);
	    xfig_line(fp, coord->x, coord->y, coord->tx, coord->ty);
	    }
	hp_b=hp_b->next; 
    }

    figtext.object=4;
    figtext.sub_type=0;
    figtext.color=0;
    figtext.depth=0;
    figtext.pen_style=0;
    figtext.font=12;
    figtext.font_size=fontsize;
    figtext.angle=0.000;
    figtext.font_flags=6;
    figtext.height=135;
    figtext.length=405;

    ha=9*fzoom;

    hp_a=dac_root.next;
    for(d=0; d<hp.nc; d++){
	figtext.sub_type=abs(hp_a->direct);
	switch(hp_a->direct){
	    case Middle_Text:
		chl=0;
		break;
	    case Left_Text:
		chl = 9*fzoom;
		break;
	    case Right_Text:
		chl = -9*fzoom;
		break;
	}
	
	for (i=0;i<strlen(hp_a->c);++i) if (hp_a->c[i]=='\\') hp_a->c[i]=' ';
if (!strpbrk(hp_a->c,"@_^")) { /*no sub- or superscripts*/
	if(latex_flag){
	    fprintf(fp, "%i %i %i %i %i %i %f %f %i %f %f %i %i $%s$\\001\n",
		figtext.object, figtext.sub_type, figtext.color,
		figtext.depth, figtext.pen_style, figtext.font,
		figtext.font_size, figtext.angle, figtext.font_flags,
		figtext.height, figtext.length,
		hp_a->x*fzoom-chl, hp_a->y*fzoom+ha, hp_a->c);
	}else{
	    fprintf(fp, "%i %i %i %i %i %i %f %f %i %f %f %i %i %s\\001\n",
		figtext.object, figtext.sub_type, figtext.color,
		figtext.depth, figtext.pen_style, figtext.font,
		figtext.font_size, figtext.angle, figtext.font_flags,
		figtext.height, figtext.length,
		hp_a->x*fzoom-chl, hp_a->y*fzoom+ha, hp_a->c);
	}
}else{ /* special formatting needed, every character becomes a separate entity*/	
	xpos=hp_a->x*fzoom-chl;
      for (i=0;i < strlen(hp_a->c);i++){
      offset=0;
      fontshrink=0;
      figtext.font=12;
      if (hp_a->c[i]=='@'){ figtext.font=32; i++;}
      if (hp_a->c[i]=='_') offset=ha/2; 
      if (hp_a->c[i]=='^') offset=-ha/2; 
      if (offset != 0) {
      fontshrink=3;
       i++;
	}
	if(latex_flag){
	    fprintf(fp, "%i %i %i %i %i %i %f %f %i %f %f %i %i $%c$\\001\n",
		figtext.object, figtext.sub_type, figtext.color,
		figtext.depth, figtext.pen_style, figtext.font,
		figtext.font_size-fontshrink, figtext.angle, figtext.font_flags,
		figtext.height, figtext.length,
		xpos, hp_a->y*fzoom+ha+offset, hp_a->c[i]);
	}else{
	    fprintf(fp, "%i %i %i %i %i %i %f %f %i %f %f %i %i %c\\001\n",
		figtext.object, figtext.sub_type, figtext.color,
		figtext.depth, figtext.pen_style, figtext.font,
		figtext.font_size-fontshrink, figtext.angle, figtext.font_flags,
		figtext.height, figtext.length,
		xpos, hp_a->y*fzoom+ha+offset, hp_a->c[i]);
	}
        xpos=xpos+chl;
      } /*for i*/
      }        
	for (i=0;i<strlen(hp_a->c);++i) if (hp_a->c[i]==' ') hp_a->c[i]='\\';
	hp_a=hp_a->next; 
    }

    fclose(fp);
    return(0);
}

int export_xfig(char *filename)
{
    char xfile[512];

    if(strlen(filename)) {
	sprintf(xfile, "%s.fig", filename);
	exfig(xfile, 0);
    }
return(0);
}

int export_latex_pic(char *filename)
{
    char com[255];
    char xfile[512];

    if(strlen(filename)) {
	sprintf(com, "fig2dev -Lpictex %s.f2l %s.tex", filename, filename);
	sprintf(xfile, "%s.f2l", filename);
	exfig(xfile, 1);
	system(com);
    }
return(0);    
}
int export_ps_pic(char *filename)
{
    char com[255];
    char xfile[512];

    if(strlen(filename)) {
	sprintf(com, "fig2dev -Lps -P %s.f2l %s.ps", filename, filename);
	sprintf(xfile, "%s.f2l", filename);
	exfig(xfile, 0);
	system(com);
    }
return(0);    
}
